#!/usr/bin/perl
# nagios: -epn

#Copyright 2013-2016 MarkLogic Corporation.  All Rights Reserved.

#Plugin to monitor MarkLogic with Nagios
#
# Nagios runs this check, by supplying the correct params
#
# ex.
# > perl check_marklogic.pl -a user:passwd -H localhost -p 8002 --path /manage/LATEST/databases
#

use strict;
use Getopt::Long qw(:config no_ignore_case);
use Nagios::Monitoring::Plugin::Threshold;
use List::Util qw[min max sum];
use Switch;
use XML::LibXML;
use POSIX qw[ _exit ];

use constant VERBOSE1 => 1;    #Main results
use constant VERBOSE2 => 2;    #Intermediate results
use constant VERBOSE3 => 3;    #Tracing
use constant VERBOSE4 => 4;    #Everything

#forces the output to happen
use constant OUTPUTTOSHELL => -1;

# based on: http://nagiosplug.sourceforge.net/developer-guidelines.html#AEN76
use constant OK          => 0;
use constant WARNING     => 1;
use constant ERROR       => 2;
use constant UNKNOWN     => 3;
use constant PLUGINCODES => qw (OK Warning Critical Unknown);

open( STDERR, ">&STDOUT" );

sub printVerbose {
	my ( $text, $vcode, $currentverboselevel ) = @_;
	printVerbose_noDefaultLineBreak( "$text \n", $vcode, $currentverboselevel );
}

# Prints output messages for debugging and co.
sub printVerbose_noDefaultLineBreak {
	my ( $text, $vcode, $currentverboselevel ) = @_;
	if ( ( $vcode <= $currentverboselevel ) || ( $vcode == OUTPUTTOSHELL ) ) {
		print "$text";
	}

}

# Wrong CMD options other major problems: script will terminate and print an error message
sub handleErrors {
	my ($errormessage) = @_;
	print $errormessage . "\n";
	exit UNKNOWN;    #e.g. when invalid arguments are specified, ...
}

# Prints the input parameters that are converted from Getop:Long
sub printInput {
	my ( $verbose, %input ) = @_;
	printVerbose( "Input-Data:", VERBOSE2, $verbose );
	foreach my $key ( keys %input ) {
		printVerbose( "-$key  --> $input{$key}", VERBOSE2, $verbose );
	}
}

# Prints the input parameters that are converted from Getop:Long
sub returnCMDOptions {
    my ( %input ) = @_;
    my $inputString = "";
    foreach my $key ( keys %input ) {
    	if($key eq "authentication"){
    		$inputString = $inputString. "[". $key . "=*****]";
    	}else{
            $inputString = $inputString. "[". $key . "=" . $input{$key} . "]";
    	}
    }
    return $inputString ;
}


#  Parses the CMD options
sub parseInput {
	my ($arguments) = @_;
	my %options = ();
	GetOptions(
		\%options,               "version|V",
		"help|h",                "verbose|v:i",
		"authentication|a=s",    "host|H=s",
		"path=s",                "port|p=i",
		"ssl:i",                 "timeout|t=i",
		"op|operator=s",         "key|k=s",
		"c=s",
		"w=s",
        "sslcertificate|cert:s",
	);
	printInput( $options{verbose}, %options );
	return %options;
}

# Checks that the required CMD options (port, machine, authentication & path) are specified.
sub checkDefaultParametersAndExtractFailoverHosts {
	my (%input) = @_;
	printVerbose( "Check-Default-parameters", VERBOSE2, $input{verbose} );

	#make sure they are there.
	if (
		$input{authentication}
		&& $input{host}
		&& $input{port}
		&& $input{path}

	  )
	{
		if ($input{authentication} =~ m/\S:\S/){
            # ok
        }else
        {
        	handleErrors(
        	"Value \"$input{authentication}\" is invalid for option authentication. Required format: username:password");
        }
		printVerbose( "-Default-parameters: specified",
			VERBOSE2, $input{verbose} );
		printVerbose(
			"-Default-parameters: authentification=$input{authentication}",
			VERBOSE3, $input{verbose} );
		printVerbose( "-Default-parameters: host=$input{host}",
			VERBOSE3, $input{verbose} );
		if($input{host} =~ m/,/){
			printVerbose( "-Host specified: $input{host}",
            VERBOSE2, $input{verbose} );
            my @temp = split(/,/ ,$input{host});
            $input{host} =  @temp[0] ;
            $input{"failoverhost"} =  @temp[1] ;

            printVerbose( "-Basic Host: $input{host}",
                VERBOSE2, $input{verbose} );
            printVerbose( "-Backup hosts: $input{failoverhost}",
                VERBOSE2, $input{verbose} );
            printVerbose( "-Backup host-count:" . @temp ,
                VERBOSE2, $input{verbose} );
            if(@temp > 2){
            	handleErrors(
            "Just one backup hosts allowed. Specified" . @temp );
            }
		}
		printVerbose( "-Default-parameters: port=$input{port}",
			VERBOSE3, $input{verbose} );
		printVerbose( "-Default-parameters: path=$input{path}",
			VERBOSE3, $input{verbose} );
	}
	else {
		handleErrors(
			"Authentication, host, port and path are required parameters");
	}
	return %input;
}

# Constructs the curl command with the help of the parameters: port, host, authentication & path
# Curl path is hard coded and can be replaced with wget & go.
sub constructCurlCommand {
	my ( $host, $port, $path, $authentication, $timeout, $ssl, $sslcertificate, $verbose ) = @_;
	printVerbose( "ConstructCurlCommand", VERBOSE3, $verbose );
    my $httpprefix = undef;
    my $cert = undef;
    if($ssl ==1){
    	 $httpprefix = "https://";
    	 if(length($sslcertificate)>0){
    	   $cert = "--cacert $sslcertificate";
    	 }
    }else{
    	 $httpprefix = "http://";
    }

    #correct invalid urls which contain two or more question marks
    #in case the url contains two or more questionsmarks, replace 2nd and following to a &
	my $seenpreviousquestionmark = 0;
	my @chars = split /(\w*)(?)(\w*)/, $path;
	for my $char(@chars){
	 if(($char eq "?") && ($seenpreviousquestionmark eq 0)){
	    $seenpreviousquestionmark = 1;
	 }elsif(($char eq "?") && ($seenpreviousquestionmark eq 1))
	 {
	 	  $char = "&";
	 	  printVerbose( "-Corrected-Path-Old: $path",
	        VERBOSE2, $verbose );
	 }
	}
    my $path = join '', @chars;
    printVerbose( "-Path: $path", VERBOSE2, $verbose );

	my $url     = $httpprefix . $host . ":" . $port . $path;
	my $maxtime = "";
	if ($timeout) {
		$maxtime = "--max-time $timeout";
	}else{
	   $maxtime = "--max-time 10";
	}
	my $entireCommand =
"curl \"$url\" $maxtime --user $authentication -i --anyauth -H \"Accept: application/xml\" --silent $cert 2>&1 ";
	printVerbose( "-Generated-Curl-Command: $entireCommand",
		VERBOSE2, $verbose );
	return $entireCommand;
}

#Check for optional arguments like -c -w
sub checkOptionalArguments {
	my (%input) = @_;
	#check for verbose
	if ( exists( $input{verbose} ) ) {
		#Note: run in verbose mode only for debugging, not within Nagios
		printVerbose(
"Attention!!! Verbose should be used for debugging purposes only! (It shouldn't be specified when called out of Nagios)",
			VERBOSE2, $input{verbose}
		);
		if ( ( 0 <= $input{verbose} ) && ( $input{verbose} <= 4 ) ) {
			printVerbose(
				"-Optional-parameter: Verbose - specified:$input{verbose}",
				VERBOSE2, $input{verbose} );
		}
		else {
			handleErrors("Value \"$input{verbose}\" is invalid for option verbose. Valid values are 0,1,2,3");
		}
	}
	else {

		# no verbose
		# can't print verbose output as verbose it not specified :-)
	}
	if ( exists( $input{ssl} ) ) {
		if ( ( 0 <= $input{ssl} ) && ( $input{ssl} <= 1 ) ) {
			printVerbose( "-Optional-parameter: SSL - specified:$input{ssl}",
				VERBOSE2, $input{verbose} );
			if ($input{ssl} == 1){
				 if ( exists ($input{sslcertificate})) {
	                printVerbose( "-Optional-parameter: SSL-Certificate - specified:$input{sslcertificate}",
	                VERBOSE2, $input{verbose} );
	             }
	             else
	             {
	             	printVerbose( "-No SSL-Certificate specified. Assumed it is available to curl with specifying --cacert ",
	                VERBOSE2, $input{verbose} );
	             }
			}
			else
			{
				if ( exists ($input{sslcertificate})) {
                    handleErrors("--sslcertificate specified but --ssl 1 is not specified. ");
                 }
			}
		}
		else {
			handleErrors("Value \"$input{ssl}\" is invalid for option ssl. Valid values are 0,1");
		}

	}
	else {
		printVerbose( "-Optional-parameter: SSL - not defined",
			VERBOSE2, $input{verbose} );
        if ( exists ($input{sslcertificate}) ) {
           printVerbose( "-Optional-parameter: SSLCertificate - specified",
            VERBOSE2, $input{verbose} );
           handleErrors("--sslCertificate specified but it requires --ssl 1. ");
        }
	}
	if ( exists( $input{key} ) ) {
		printVerbose( "-Optional-parameter: KEY - specified:$input{key}",
			VERBOSE2, $input{verbose} );
		if ( $input{key} !~ /[a-z]/ ) {
			printVerbose( "Value \"$input{key}\" should most likely be typed in lower case",
				VERBOSE2, $input{verbose} );
		}
		if(exists($input{c}) or exists($input {w})){
			printVerbose( "-Optional-parameter: KEY - has -c or -w",
            VERBOSE2, $input{verbose} );
		}else
		{
			printVerbose( "Optional Parameter --key specified but no thresholds with -c or -w specified",
            VERBOSE2, $input{verbose} );
		}

	}
	else {
		printVerbose( "-Optional-parameter: Key - not defined",
			VERBOSE2, $input{verbose} );
	}
	if ( exists( $input{timeout} ) ) {
		printVerbose(
			"-Optional-parameter: Timeout - specified:$input{timeout}",
			VERBOSE2, $input{verbose} );
		if ( 0 <= $input{timeout} ) {
			printVerbose( "-Optional-parameter: Timeout - ok",
				VERBOSE2, $input{verbose} );
		}
		else {
			handleErrors("Value \"$input{timeout}\" is invalid for option timeout. Timeout has to be a positive integer");
		}
	}
	else {
		printVerbose( "-Optional-parameter: Timeout - not defined",
			VERBOSE2, $input{verbose} );
	}
	if ( ( exists( $input{c} ) ) || ( exists( $input{w} ) ) ) {
		printVerbose(
			"-Optional-parameter: -w OR -c - specified:$input{w}-$input{c}",
			VERBOSE2, $input{verbose} );

# if the specified arguments of -c and -w are valid, the set_threshold function accepts them.. if not.. it terminates
		my $threshold = Nagios::Monitoring::Plugin::Threshold->set_thresholds(
			warning  => "$input{w}",
			critical => "$input{c}"
		);
        if ( not $input{key} ) {
			handleErrors("-c or -w requires --key");
		}
	}
	else {
		printVerbose( "-Optional-parameter: -w OR -c - not specified",
			VERBOSE2, $input{verbose} );
	}
	if ( exists( $input{op} ) ) {
		printVerbose( "-Optional-parameter: op - specified:$input{op}",
			VERBOSE2, $input{verbose} );
		if (   ( $input{op} eq "range" )
			|| ( $input{op} =~ m/^(eq|ne)=/ ) )
		{
			printVerbose( "-Optional-parameter: op: Test passed",
				VERBOSE2, $input{verbose} );
			if ( ( exists( $input{c} ) ) || ( exists( $input{w} ) ) ) {
				printVerbose( "-Optional-parameter: op: with -w OR -c",
                VERBOSE2, $input{verbose} );
			}
			else
			{
				handleErrors("-op requires -w OR -c");
			}
		}
		else {
			handleErrors("Value \"$input{op}\" is invalid for option \"op\". Possible values are eq|ne|range");
		}
	}
}

# Takes a nodelist as input. Returns the values in an array
sub generateValueList {
	my ( $verbose, @nodelist ) = @_;
	my $value = "";
	my $unique_name = "";
	my @valuelist   = ();
	foreach my $node (@nodelist) {
		#if (  $node->hasChildNodes() ) {
			my $value = $node->textContent;
			printVerbose( "-Key-ValuePair: $value", VERBOSE4, $verbose );
			push(@valuelist, $value)
	    #}
	}

	printVerbose( "-ValueList: @valuelist", VERBOSE2, $verbose );
	return \@valuelist;
}

# Calculates the return code of this plugin call
sub determinePluginReturnCode {
	my ( $tree,$curl, %input ) = @_;
	my $getpluginthresholdcode  = "";
	my $getpluginresponseMetric = "";
	my $nagiosStatusMessage      = "";

	printVerbose( "-determinePluginReturnCode", VERBOSE3, $input{verbose} );
	my $fullpath = "";

	my $root     = $tree->getDocumentElement;
	my $xpc      = XML::LibXML::XPathContext->new($root);
	my $node     = ( $xpc->findnodes('/*') )[0];
	my @nslist   = $node->getNamespaces();
	my $ns       = @nslist[0];
	my $nsuri    = $ns->getValue();
	#return the namespace prefix (even when called getLocalName as it is an instance of XML::LibXML::Namespace
	#my $nsprefix = $ns->getLocalName();

	printVerbose( "-reducing data to key: $input{key}",
	   VERBOSE3, $input{verbose} );

	$fullpath = "//*[local-name(.)= \"$input{key}\" ]";
	printVerbose( "-Xpath: $fullpath",
	   VERBOSE3, $input{verbose} );


	my $unique_name = "";
	my @nodelist    = $xpc->findnodes($fullpath);

	my ($two_ref) = generateValueList( $input{verbose}, @nodelist );
	my @valuelist = @$two_ref;

#second evaluate -op on all of the remaining key-value pairs in $reducedCurlResponse
	if ( ( $input{op} eq "range" ) || !( $input{op} ) ) {

		printVerbose( "-eval-Range/Default-Range", VERBOSE3, $input{verbose} );
		if ( $input{c} ) {
			printVerbose( "-found-c-Range", VERBOSE3, $input{verbose} );
		}
		if ( $input{w} ) {
			printVerbose( "-found-w-Range", VERBOSE3, $input{verbose} );
		}
		my $threshold = Nagios::Monitoring::Plugin::Threshold->set_thresholds(
			warning  => "$input{w}",
			critical => "$input{c}"
		);

		#is there is a single value
		if ( (@valuelist) == 1 ) {
			printVerbose(
"-Single input: @valuelist[0]",
				VERBOSE3, $input{verbose}
			);
			if ( @valuelist[0] =~ m/^(\d+\.?\d*|\.\d+)$/ ) {
				# use get_status to evaluate the value against -c -w ranges
				$getpluginthresholdcode =
				  $threshold->get_status( @valuelist[0] );
				  printVerbose( "-Threshold: $getpluginthresholdcode", VERBOSE3, $input{verbose} );
				$nagiosStatusMessage = @valuelist[0];
			}
			elsif (length(@valuelist[0]) eq 0) {
				handleErrors("The filter --key " . $input{key} . " returned an empty element. Check the API.");
			}
			else {
                handleErrors("Expected numeric data. Use -eq or -ne to process string input: @valuelist ");
            }
		}
		elsif (scalar(@valuelist) >1 )
		{
			handleErrors("API returned more than one value for the key($input{key}). This is not supported.: @valuelist");
		}
		else
		{
			handleErrors("The XML element ($input{key}) was not present in the data returned by the API ($curl)");
		}
	}
	elsif ( ( $input{op} =~ m/eq=/ ) or ( $input{op} =~ m/ne=/ ) ) {
		printVerbose( "-Matching  $input{op}", VERBOSE2, $input{verbose} );
		#input  eq=open ne=aborted,...
		if ( ( $input{op} =~ m/eq=(.*)/ ) || ( $input{op} =~ m/ne=(.*)/ ) ) {
			my $matching = "$1";
			printVerbose( "-EQ-to match:'$matching'", VERBOSE3,
				$input{verbose} );
			my @boollist = ();
			foreach my $myvalue (@valuelist) {
				if ( $input{op} =~ m/eq=/ ) {
					#eq case
					if ( $myvalue eq $matching ) {
						@boollist = ( @boollist, 1 );
					}
					else {
                		@boollist = ( @boollist, 0 );
					}
				}
				else {
					#ne case
					if ( $myvalue eq $matching ) {
                		@boollist = ( @boollist, 0 );
					}
					else {
						@boollist = ( @boollist, 1 );
					}
				}
			}
			printVerbose( "-EQ-NE Temp Values: @boollist",
				VERBOSE3, $input{verbose} );
			if ( @boollist == 1 ) {
				printVerbose( "-SingleValue:$nagiosStatusMessage",
					VERBOSE2, $input{verbose} );
				$nagiosStatusMessage = @boollist[0];

				if ( ( $input{w} ) or ( $input{c} ) ) {
					my $threshold = Nagios::Monitoring::Plugin::Threshold->set_thresholds(
						warning  => "$input{w}",
						critical => "$input{c}"
					);
					$getpluginthresholdcode =
					  $threshold->get_status($nagiosStatusMessage);

				}
				else {
					$getpluginthresholdcode = UNKNOWN;
				}
			}
			elsif (@boollist > 1) {
			    handleErrors("API returned more than one value for the key($input{key}). This is not supported: @boollist");
			}
			else {
				handleErrors("XML element ($input{key}) was not present in the data returned by the API");
			}

		}
	}
	else {
		handleErrors(
"Invalid -op option. Valid options are ne|ne|range"
		);
	}
	printVerbose( "-NagiosPluginReturnCode:$getpluginthresholdcode",
		VERBOSE3, $input{verbose} );

	my $returnCodeAsString = (PLUGINCODES)[$getpluginthresholdcode];
	#adds the warning and critical threshold value into the performance element for the matching value
	my $thresholdNote = "";
	if($input{c}){
		$thresholdNote  = "[critical($input{c})]";
	}
	if($input{w}){
        $thresholdNote = $thresholdNote .  "[warning($input{w})]";
    }
    printVerbose( "-NagiosAdditionalPrintingOfThreshold:$thresholdNote",
        VERBOSE3, $input{verbose} );

	$getpluginresponseMetric =
	  "$returnCodeAsString - $input{key}=$nagiosStatusMessage $thresholdNote ";
	my $additionalPerfData = "";

	return $getpluginthresholdcode, $additionalPerfData,
	  $getpluginresponseMetric;

}

# Returns the output in the 'label'=value[UOM];
# Returns performance metrics for all elements which have an attribute "Units"
sub getPerfData {
	my ( $tree, $additionalPerfData,$key,$w,$c, $verbose ) = @_;
	printVerbose(
"-getPerfData: Extracts from XML all boolean and numeric values which have an \"units\" attribute",
		VERBOSE4, $verbose
	);
	my $output = "";
	if($additionalPerfData)
	{
		$output = " $additionalPerfData;";
	}

	my $root = $tree->getDocumentElement;
	my $xpc  = XML::LibXML::XPathContext->new($root);

    my $fullpath = undef;
    my $unique_name = undef;

    $fullpath = "//*[local-name(.) = \"name\"]";
    my @nodelist = $xpc->findnodes($fullpath);
    my $countOfNameNodes = @nodelist;

    printVerbose("-countOfNameNodes:" . $countOfNameNodes,
               VERBOSE4, $verbose );

    if(length($key)>0){
        # extract all elements which are below or equal to the key
        $fullpath    = "//*[local-name(.)= \"$key\" or local-name(.)= \"name\" ]/descendant-or-self::node()[\@units or local-name(.)= \"name\" or local-name(.)= \"$key\"] ";
    }
    else
    {
        #no key. use all elements with units attribute.
    	$fullpath  = "//*[\@units or local-name(.)= \"name\"]";
    }
    my @nodes = undef;
    @nodes = $xpc->findnodes($fullpath);

    printVerbose("-number of Nodes for Perf data:" . @nodes,
               VERBOSE4, $verbose );
    if((@nodes == $countOfNameNodes)){
        printVerbose("--key \'$key\' did not match any elements in the view. Use --verbose 1 to debug it.",
               VERBOSE2, $verbose );
    }
    if(length($c)<1 && length($w)<1 && (@nodes == $countOfNameNodes) )
    {
    	handleErrors("--key \'$key\' did not match any elements in the view. Use --verbose 1 to debug it.");
    }
	foreach my $node (@nodes ) {
		my $threshold = undef;
		my $temp = $node->getLocalName;
		if($temp eq $key){
			if($w){
				$threshold = ";$w;";
			}
			else{
				$threshold = ";;";
			}
			if($c){
			  $threshold = $threshold. "$c;";
			}

		}else
		{
			$threshold = ";";
		}
		if ( $temp =~ m/name/ ) {
			$unique_name = $node->textContent;
			if(length($unique_name) != 0){
				$unique_name = $unique_name . "-";
			}
			printVerbose( "-unique-Name found: $unique_name",
				VERBOSE4, $verbose );
		}
		if ( $node->hasAttribute('units') ) {
			my $uom   = $node->getAttribute("units");
			my $value = $node->textContent;
             printVerbose("-Value: $uom - $value -",
               VERBOSE4, $verbose );
			switch ($uom) {
				case "bool" {
                    $uom =
                      "";  #Todo: for boolean we can add the min max things too.
                    if ( $value eq "true" )
                       {$value = 1; }
                    else { $value = 0 }
                }
				case "B" { }
				case "MB" { }
				case "GB" { }
				case "TB" { }
				case "sec"   { $uom = "s"}
				case "%" { }
				case "enum" {}
                else { $uom = "";
					# this filters e.g. the unit of mesure "count"
				}
			}
			if(length($value) ==0){
				$value = 0;
			}
			#filter units with "enum" as they are likely some text
			if($uom eq "enum") {} else {
			 my $perfelement = "$unique_name$temp=$value$uom$threshold";
			 printVerbose( "-PerfCounter: $perfelement", VERBOSE2, $verbose );
			 $output = "$output $perfelement";
			}
		}elsif($node->getLocalName() eq $key){
			 if($node->textContent =~ m/^(\d+\.?\d*|\.\d+)$/)
			 {
				my $value = $node->textContent;
				my $perfelement = "$unique_name$temp=$value;$threshold";
				$output = "$output $perfelement";
			 }elsif($node->textContent eq "true"){
				my $perfelement = "$unique_name$temp=1;$threshold";
                $output = "$output $perfelement";
			 }elsif ($node->textContent eq "false")
			 {
				my $perfelement = "$unique_name$temp=0;$threshold";
                $output = "$output $perfelement";
			 }
			 else{
			 	printVerbose( "Value of XML element \'$key\' doesn't look like a numeric value:" . $node->getLocalName(), VERBOSE2, $verbose );
			 }

		}
	}

	printVerbose( "-completePerfData: $output", VERBOSE2, $verbose );
	return $output;
}

sub validateMonitoringAPIResultandReduceToXML {
	my ( $curl_command, $returnfromCurl, $verbose ) = @_;
	my $message = "";

	printVerbose( "-Return-from-Curl: $returnfromCurl", VERBOSE2, $verbose );

	if($returnfromCurl =~ m/Connection: (keep-alive|close)/m)
	{
		printVerbose( "-curl-reponse basic test true", VERBOSE2, $verbose );
	}


	elsif ( $returnfromCurl =~ m/certificate/m ) {
        $message = "SSL certificate problem -  $curl_command - $returnfromCurl  ";
	}

    # split and isolate header of interest
    my @temp = (split(/^\s*$/m, $returnfromCurl, 3));
    $returnfromCurl= @temp[1];

    if ( $returnfromCurl =~ m/HTTP\/1.1 200/ ) {
        if ( $returnfromCurl =~ m/Content-Length: 0/ ) {
            $message = ("Management-API: returned: Content-Length: 0 - $curl_command ");
        }
        $returnfromCurl = @temp[2];
        $returnfromCurl =~ s/^\s+//;
    }
    elsif ( $returnfromCurl =~ m/HTTP\/1.1 400/ ) {
		$message = ("Bad Request (400): $curl_command");
	}
	elsif ( $returnfromCurl =~ m/HTTP\/1.1 401/ ) {
        $message = ("Unauthorized (401): $curl_command");
    }
	elsif ( $returnfromCurl =~ m/HTTP\/1.1 403/ ) {
        $message = ("Forbidden (403): $curl_command");
    }
	elsif ( $returnfromCurl =~ m/HTTP\/1.1 404 Not Found/ ) {
		$message = ("Not Found (404): $curl_command");
	}elsif ($returnfromCurl =~ m/undefined error/){
		$message = ("Not Found: $curl_command");
	}elsif ($returnfromCurl =~ m/Resource or view does not exist/){
        $message = ("Not Found: $curl_command");
	}elsif ($returnfromCurl =~ m/error:error/){
		$message = ("Not Found: $curl_command");
	}elsif ($returnfromCurl =~ m/HTTP\/1.1 503/){
        $message = ("Management-API disabled: $curl_command");
    }elsif ( $returnfromCurl =~ m/HTTP\/1.1 401/ ) {
        $message = ("Unauthorized: $curl_command");
    }elsif ( $returnfromCurl =~ m/Content-Length: 0/ ) {
        $message = ("Management-API: returned: Content-Length: 0 -  $curl_command ");
    }
	elsif ( $returnfromCurl eq "" ) {
		$message = ("Not Found (404): $curl_command"
		);
	}
	else{
		$message = ("Management-API: returned: $curl_command (not HTTP/1.1 200) ");
	}

	return ($returnfromCurl, $message);
}

sub doThresholding {
	my ( $tree,$curl, %input ) = @_;

	my $PluginReturnCode     = UNKNOWN;
	my $PluginresponseMetric = "";

	my $PerfMetric = "";
	printVerbose( "-doThresholding", VERBOSE3, $input{verbose} );
	if ( ( $input{w} ) or ( $input{c} ) ) {
		( $PluginReturnCode, $PerfMetric, $PluginresponseMetric ) =
		  determinePluginReturnCode( $tree, $curl, %input );
	}
	else {
		$PluginresponseMetric = "OK ";
		$PluginReturnCode     = OK;
		printVerbose( "-PluginresponseMetric:$PluginresponseMetric", VERBOSE2, $input{verbose} );
	}
	return ( $PluginReturnCode,$PerfMetric, $PluginresponseMetric );
}

# Step 1
# Parse Input args
my %inputArgs = ();
local $@;
eval {
	local $SIG{__WARN__};

	#uses Getop:Long
	%inputArgs = parseInput($ARGV);
};

# Step 2
# Check if user specified 'help' or 'version' or 'asked' for normal program execution
if ( exists $inputArgs{version} ) {
	printVerbose( "In-Version",    VERBOSE3,      $inputArgs{verbose} );
	printVerbose( "check_marklogic.pl: version 1.0.1", OUTPUTTOSHELL, $inputArgs{verbose} );
	exit UNKNOWN;
}
elsif ( exists $inputArgs{help} ) {
	printVerbose( "Starting Help", VERBOSE3,$inputArgs{verbose} );
	printVerbose( "Help - check_marklogic.pl",OUTPUTTOSHELL, $inputArgs{verbose} );
	printVerbose( "-----------------------------",OUTPUTTOSHELL, $inputArgs{verbose} );
	printVerbose( "This Nagios Plugin monitors MarkLogic Server.",OUTPUTTOSHELL, $inputArgs{verbose});
	printVerbose( "It retrieves performance data from the API, applies thresholds (when specified) and adds additional performance data to the output",OUTPUTTOSHELL, $inputArgs{verbose} );
    printVerbose( "-----------------------------",OUTPUTTOSHELL, $inputArgs{verbose} );
	    printVerbose( "-a (--authentication) <username:password>",OUTPUTTOSHELL, $inputArgs{verbose} );
	    printVerbose( "-H (--host)           <hostname> ",OUTPUTTOSHELL, $inputArgs{verbose} );
    printVerbose( "-p (--port)           <port>             (e.g. 8002)",OUTPUTTOSHELL, $inputArgs{verbose} );
    printVerbose( "-path (--path)        <path>             (e.g. /manage/LATEST/forests/Documents?view=status)",OUTPUTTOSHELL, $inputArgs{verbose} );
    printVerbose( "---------+ Optional parameters for Thresholding------------------",OUTPUTTOSHELL, $inputArgs{verbose} );
    printVerbose( "-k (--key)           <string>            (e.g. device-space",OUTPUTTOSHELL, $inputArgs{verbose} );
    printVerbose( "-op (--operator)     <(range|(eq|ne)=string))> (default range)",OUTPUTTOSHELL, $inputArgs{verbose} );
    printVerbose( "-w (--critical)      <range>             (default not set)",OUTPUTTOSHELL, $inputArgs{verbose} );
    printVerbose( "-c (--critical)      <range>             (default not set)",OUTPUTTOSHELL, $inputArgs{verbose} );
    printVerbose( "---------+ Optional parameters for establishing the connection------------------",OUTPUTTOSHELL, $inputArgs{verbose} );
    printVerbose( "-ssl (--ssl)          <0|1> (default 0)",OUTPUTTOSHELL, $inputArgs{verbose} );
    printVerbose( "-cert (--sslcertificate)          <fullpath> (requires -ssl 1)",OUTPUTTOSHELL, $inputArgs{verbose} );

    printVerbose( "-t (--timeout)        <timeout-seconds>  (e.g. 5, default 10) ",OUTPUTTOSHELL, $inputArgs{verbose} );

    printVerbose( "",OUTPUTTOSHELL, $inputArgs{verbose} );
    printVerbose( "---------OR------------------",OUTPUTTOSHELL, $inputArgs{verbose} );
    printVerbose( "-h (--help)",OUTPUTTOSHELL, $inputArgs{verbose} );
    printVerbose( "-V (--version) ",OUTPUTTOSHELL, $inputArgs{verbose} );


	printVerbose( "-verbose 0|1|2|3                        (default 0)",OUTPUTTOSHELL, $inputArgs{verbose} );
	printVerbose( "    (Use --verbose 2|3 for debugging from Command Line only)",OUTPUTTOSHELL, $inputArgs{verbose} );
	printVerbose( "     -verbose 0: No debug output",OUTPUTTOSHELL, $inputArgs{verbose} );
    printVerbose( "     -verbose 1: Adds the user input in the status message. Good for debugging of Nagios configuration files",OUTPUTTOSHELL, $inputArgs{verbose} );
    printVerbose( "     -verbose 2: Adds limited debug messages. Best for command line debugging",OUTPUTTOSHELL, $inputArgs{verbose} );
    printVerbose( "     -verbose 3: Adds all debug messages. Best for command line debugging",OUTPUTTOSHELL, $inputArgs{verbose} );
    printVerbose( "",OUTPUTTOSHELL, $inputArgs{verbose} );
    printVerbose( "Examples:",OUTPUTTOSHELL, $inputArgs{verbose} );
    printVerbose( "perl check_marklogic.pl -a user:pw -H localhost -p 8002 -path /manage/LATEST",OUTPUTTOSHELL, $inputArgs{verbose} );
    printVerbose( "perl check_marklogic.pl -a user:pw -H localhost -p 8002 -path /manage/LATEST/forests/Documents?view=status --key state  -op eq=open -c @0:0",OUTPUTTOSHELL, $inputArgs{verbose} );
    printVerbose( "perl check_marklogic.pl -a user:pw -H localhost -p 8002 -path /manage/LATEST/forests?view=status --key total-forests -c @1:200 -w @21:200",OUTPUTTOSHELL, $inputArgs{verbose} );

	exit UNKNOWN;
}
else {
	my $PluginReturnCode = UNKNOWN;

	#make sure a user specified all the default parameters
	printVerbose( "Starting Main Execution", VERBOSE3, $inputArgs{verbose} );
	%inputArgs = checkDefaultParametersAndExtractFailoverHosts(%inputArgs);

	#make sure verbose & co have valid parameters
	printVerbose( "-checkOptionalArguments", VERBOSE3, $inputArgs{verbose} );
	checkOptionalArguments(%inputArgs);

	#construct curl command
	my $curl_command = undef;

    my $curl_succeded = 0;
    my $message = "";
    my @hosts = ();
    push(@hosts, $inputArgs{host});
    if($inputArgs{failoverhost}){
        push(@hosts, $inputArgs{failoverhost});
    }
    my $i = 0;
    my $host = @hosts[$i];
    my $returnFromCommonAPI = undef;
    my $failover = 0;

    while ($host && $curl_succeded == 0){

       #failover loop start
	   $curl_command= constructCurlCommand(
	        $host,    $inputArgs{port},
	        $inputArgs{path},    $inputArgs{authentication},
	        $inputArgs{timeout}, $inputArgs{ssl}, $inputArgs{sslcertificate}, $inputArgs{verbose}
	         );

		#excute curl-command
		$returnFromCommonAPI = `$curl_command`;
		# do not show the username and pw in any output, also remove stderr-redirect
		$curl_command  =~ s/($inputArgs{authentication})/[username:pw]/;
		$curl_command  =~ s/2>\/dev\/null//;


		($returnFromCommonAPI, $message) = validateMonitoringAPIResultandReduceToXML( $curl_command, $returnFromCommonAPI,
			$inputArgs{verbose} );
		if(length($message)> 1){
			if($inputArgs{failoverhost}){
				 printVerbose("Warning: Http request to host $inputArgs{host} failed with: $message" , VERBOSE2, $inputArgs{verbose} );
				 $failover = 1;
			}else{
			     handleErrors($message);
			}
	    }else
	    {
	    	printVerbose("-curl_succeded" , VERBOSE2, $inputArgs{verbose} );
	    	$curl_succeded = 1;
	    }
	 if($curl_succeded ==0){
	   $i = $i +1;
	   $host = @hosts[$i];
	 }
    #failover loop ends here.
    }
    if ($curl_succeded == 0){
    	handleErrors($message);
    }else
    {
    	printVerbose("-curl succeded on host $host: ($returnFromCommonAPI)" , VERBOSE2, $inputArgs{verbose} );
    }
	my $parser = XML::LibXML->new();
	my $tree   = $parser->parse_string($returnFromCommonAPI);

	#process-input-args
	my $Key      = "";
	my $PerfData = "";

	( $PluginReturnCode,my $PerfMetric, $Key ) = doThresholding( $tree,$curl_command, %inputArgs );

    # if --key specified: limit the output to that subtree.

	$PerfData = getPerfData( $tree,$PerfMetric,$inputArgs{key},$inputArgs{w},$inputArgs{c}, $inputArgs{verbose},  );

	#enhance the status message for the Nagios overview panels
    if ( ( $inputArgs{w} ) or ( $inputArgs{c} ) )
    {
    	# nothing as we have already a nice message
    }else
    {
    	my @temp = split(/; /,$PerfData);
    	$Key = $Key . "-" . @temp[0] . " ";
    	$Key =~ s/;;//;
    }

    if ($failover == 1 ){
    	$Key = "Warning - used backup host. Plugin response: ($Key)";
    }
    printVerbose( "-finalStep-ReturnCode:$PluginReturnCode",
        VERBOSE2, $inputArgs{verbose} );
    printVerbose( "-finalStep-Key:$Key", VERBOSE2, $inputArgs{verbose} );
    printVerbose( "-finalStep-PerfData:$PerfData \n",
        VERBOSE2, $inputArgs{verbose} );

	#final output generation:
	if($inputArgs{verbose}==VERBOSE1){
		printVerbose_noDefaultLineBreak( $Key . "(Verbose=1 -->" . returnCMDOptions(%inputArgs) . ")|" . $PerfData . "\n",
        OUTPUTTOSHELL, $inputArgs{verbose} );
	}else{
	   printVerbose_noDefaultLineBreak( $Key . "|" . $PerfData . "\n",
		OUTPUTTOSHELL, $inputArgs{verbose} );
	}

    exit($PluginReturnCode);
}
