#!/usr/bin/perl
#Copyright 2013-1016 MarkLogic Corporation.  All Rights Reserved.

#Script to generate a Nagios config file for a MarkLogic cluster.
#It has to be used it combination with a generic config file called, 'ML-Commands-Generic.cfg'
#General settings can be changed by modifying the generic config file.
use strict;
use Getopt::Long qw(:config no_ignore_case);
use Nagios::Monitoring::Plugin::Threshold;
use List::Util qw[min max sum];
use Switch;
use XML::LibXML;
use List::MoreUtils qw/ uniq /;

use constant ML7_CONFIG_TEMPLATE => "ml_v7_template.xml";
use constant ML56_CONFIG_TEMPLATE => "ml_v5+6_template.xml";

use constant VERBOSE1 => 1;    #No Tracing


use constant VERBOSE1 => 1;    #No Tracing
use constant VERBOSE2 => 2;    #
use constant VERBOSE3 => 3;    #More Tracing
use constant VERBOSE4 => 4;    #Everything

#forces the output to happen
use constant OUTPUTTOSHELL => -1;

#constants used to extract the type of the config-templates in the xml input file
use constant CLUSTER  => "Local-Cluster";         #Main results
use constant APPSERS  => "Servers";         #Intermediate results
use constant DBS      => "Databases";       #Tracing
use constant HOSTS    => "Hosts";           #Everything
use constant TAS      => "Transactions";    #Everything
use constant REQUESTS => "Requests";        #Everything
use constant GROUPS   => "Groups";          #Everything
use constant CLUSTERS   => "Foreign-Clusters";      #Everything

use constant WARNING  => "warning";
use constant CRITICAL => "critical";

use constant DBCOUNT => "\$DatabasesCount";
use constant SERVERCOUNT=> "\$ServerCount";
use constant HOSTCOUNT=> "\$HostCount";
use constant CLUSTERCOUNT=> "\$ClusterCount";

use constant RESOURCETYPE => "resource-type";
use constant ALIAS => "alias";
use constant SERVICEDESCRIPTION => "service-description";

my $apiversion;

#Todo
# add checking that -w and -c is true for all services

open( STDERR, ">&STDOUT" );

# Stores the names of the XML elements which specify the resource-names for a given resource type
# To add custom resource-types, modify this structure.
sub getLinkSequence {
	my $linkSequence = "
	<links>
	   <link resourcetype=\"Databases\" apiName=\"Databases\" iterate=\"true\">uriref</link>
	   <link resourcetype=\"Forests\" apiName=\"Forests\" iterate=\"true\">uriref</link>
       <link resourcetype=\"Hosts\" apiName=\"Hosts\" iterate=\"true\">uriref</link>
       <link resourcetype=\"Servers\" apiName=\"Servers\" iterate=\"true\">uriref</link>
       <link resourcetype=\"Local-Cluster\" apiName=\"Local-Cluster\" iterate=\"false\">uriref</link>
       <link resourcetype=\"Foreign-Clusters\" apiName=\"Clusters\" iterate=\"true\" resource-filter=\"?cluster-role=foreign\">uriref</link>
    </links>
    ";
	return $linkSequence;

}

sub printVerbose {
	my ( $text, $vcode, $currentverboselevel ) = @_;
	printVerbose_noDefaultLineBreak( "$text \n", $vcode, $currentverboselevel );
}

# Prints output messages for debugging and co.
sub printVerbose_noDefaultLineBreak {
	my ( $text, $vcode, $currentverboselevel ) = @_;
	if ( ( $vcode <= $currentverboselevel ) || ( $vcode == OUTPUTTOSHELL ) ) {
		print "$text";
	}

}

# Wrong CMD options or other major problems: script will terminate and print an error message
sub handleErrors {
	my ($errormessage) = @_;
	print $errormessage . "\n";
	exit -1;    #e.g. when invalid arguments are specified, ...
}

# Prints the input parameters that are converted from Getop:Long
sub printInput {
	my ( $verbose, $leadingChar, %input ) = @_;
	printVerbose(
        $leadingChar . "Note: It is recommended to generate the Nagios configuration for MarkLogic with \"--verbose 1\"",
        OUTPUTTOSHELL, $verbose );
    printVerbose(
		$leadingChar . "This configuration file was generated with these parameters:",
		OUTPUTTOSHELL, $verbose );

	foreach my $key ( keys %input ) {
		if ( $key eq 'authentication' ) {
			printVerbose( $leadingChar . "-authentication  user:pw",
				OUTPUTTOSHELL, $verbose );
		}
		else {
			printVerbose( $leadingChar . "-$key  --> $input{$key}",
				OUTPUTTOSHELL, $verbose );
		}
	}
}

# Parses the CMD options
sub parseInput {
	my ($arguments) = @_;
	my %options = ();
	GetOptions(
		\%options,               "version|V",
		"help|h",                "verbose|v:i",
		"authentication|a=s",    "host|H=s",
		"port|p=i",              "filename|f=s",
		"ssl:i",                 "sslcertificate|cert:s",
		"timeout|t=i",
		"clustername|c=s",       "uniqueshortcut|u=s",
		"backuphost|bh=s",       "adminuserpw|admin=s",
	);
	return %options;
}

# Checks that an input paramter has word characters A-Z and a-z and - and _ only.
sub validInputForName{
 my ($name,$value) = @_;
 if ($value =~ m/[^a-zA-Z0-9_-]/){
    handleErrors("$name can contain \"word\" characters, \"-\" and \"_\" only");
 }
}

# Checks that the required CMD options (port, machine, authentication & path) are specified.
sub checkDefaultParameters {
	my (%input) = @_;
	printVerbose( "Check-Default-parameters", VERBOSE2, $input{verbose} );

	#make sure they are there.
	if (   $input{authentication}
		&& $input{host}
		&& $input{port}
		&& $input{clustername}
		&& $input{uniqueshortcut})
	{
		if ( $input{authentication} =~ m/\S:\S/ ) {
			#ok
		}
		else {
			handleErrors(
"Value \"$input{authentication}\" invalid for option authentication. Required format: username:password"
			);
		}

		#validInputForName("Clustername",$input{clustername});
		validInputForName("Uniqueshortcut",$input{uniqueshortcut});

		printVerbose(
			"-Default-parameters: authentification=$input{authentication}",
			VERBOSE3, $input{verbose} );
		printVerbose( "-Default-parameters: host=$input{host}",
			VERBOSE3, $input{verbose} );
		printVerbose( "-Default-parameters: port=$input{port}",
			VERBOSE3, $input{verbose} );
		printVerbose( "-Default-parameters: clustername=$input{clustername}",
            VERBOSE3, $input{verbose} );
		printVerbose( "-Default-parameters: uniqueshortcut=$input{uniqueshortcut}",
            VERBOSE3, $input{verbose} );
    }
	else {
		handleErrors(
"Authentication, host, port, clustername and uniqueshortcut are required parameters. Check --help for more information."
		);

	}

}

# Check for optional arguments
sub checkOptionalParameters{
	my (%input) = @_;
    printVerbose( "-Check-Optional-parameters", VERBOSE2, $input{verbose} );
	if ( exists( $input{ssl} ) ) {
        if ( ( 0 <= $input{ssl} ) && ( $input{ssl} <= 1 ) ) {
            printVerbose( "-Optional-parameter: SSL - specified:$input{ssl}",
                VERBOSE2, $input{verbose} );
            if ($input{ssl} == 1){
                 if ( exists ($input{sslcertificate})) {
                    printVerbose( "-Optional-parameter: SSL-Certificate - specified:$input{sslcertificate}",
                    VERBOSE2, $input{verbose} );
                 }
                 else
                 {
                    printVerbose( "-No SSL-Certificate specified. Assumed it is available to curl with specifying --cacert ",
                    VERBOSE2, $input{verbose} );
                 }
            }
            else
            {
                if ( exists ($input{sslcertificate})) {
                    handleErrors("--sslcertificate specified but --ssl 1 is not specified. ");
                 }
            }
        }
        else {
            handleErrors("Value \"$input{ssl}\" is invalid for option ssl. Valid values are 0,1");
        }

    }
    else {
        printVerbose( "-Optional-parameter: SSL - not defined",
            VERBOSE2, $input{verbose} );
        if ( exists ($input{sslcertificate}) ) {
           printVerbose( "-Optional-parameter: SSLCertificate - specified",
            VERBOSE2, $input{verbose} );
           handleErrors("--sslCertificate specified but it requires --ssl 1. ");
        }
    }
	if($input{backuphost}){
		printVerbose( "-Optional-parameters: backuphost=$input{backuphost}",
            VERBOSE3, $input{verbose} );
	}
    if ( exists( $input{timeout} ) ) {
        printVerbose(
            "-Optional-parameter: Timeout - specified:$input{timeout}",
            VERBOSE2, $input{verbose} );
        if ( 0 <= $input{timeout} ) {
            printVerbose( "-Optional-parameter: Timeout - ok",
                VERBOSE2, $input{verbose} );
        }
        else {
            handleErrors("Value \"$input{timeout}\" is invalid for option timeout. Timeout has to be a positive integer");
        }
    }
    else {
        printVerbose( "-Optional-parameter: Timeout - not defined",
            VERBOSE2, $input{verbose} );
    }
}

# Constructs the curl command with the help of the parameters: port, host, authentication & path
# Curl path is hard coded and can be replaced with wget & go.
sub constructCurlCommand {
	my ( $host, $port, $path, $authentication, $timeout, $ssl,$sslcertificate, $verbose ) = @_;
	printVerbose( "-constructCurlCommand", VERBOSE3, $verbose );
	my $httpprefix = undef;
	my $cert = undef;
	if ( $ssl == 1 ) {
		$httpprefix = "https://";
		if(length($sslcertificate)>0){
           $cert = "--cacert $sslcertificate";
        }
	}
	else {
		$httpprefix = "http://";
	}
	my $url     = $httpprefix . $host . ":" . $port . $path;
	my $maxtime = "";
	if ($timeout) {
		$maxtime = "--max-time $timeout";
	}
	else {
		$maxtime = "--max-time 10";
	}
	my $entireCommand =
"curl \"$url\" $maxtime --user $authentication -i --anyauth -H \"Accept: application/xml\" --silent $cert 2>&1 ";
	printVerbose( "-generated-Curl-Command: $entireCommand",
		VERBOSE3, $verbose );
	return $entireCommand;
}

# checks the HTTP headers.
# the second header contains the actual header
sub checkheadersandreturnXML {
	my ( $returnFromCommonAPI, $verbose ) = @_;
	printVerbose( "-Curl response: $returnFromCommonAPI",
		VERBOSE3, $verbose );

	if($returnFromCommonAPI=~ m/certificate subject name/m){
        handleErrors(
            "SSL certificate problem:$returnFromCommonAPI"
        );
	}

    my @temp = (split(/^\s*$/m, $returnFromCommonAPI, 3));
	$returnFromCommonAPI = @temp[1];


	if ( $returnFromCommonAPI =~  m/HTTP\/1.1 200/ ) {
		printVerbose( "-HTTP return code:OK(200)", VERBOSE3, $verbose );
		$returnFromCommonAPI = @temp[2];
		$returnFromCommonAPI =~ s/^\s+//;
		printVerbose( "-remaining response from curl: $returnFromCommonAPI",
			VERBOSE3, $verbose );
		return $returnFromCommonAPI;
	}
	elsif($returnFromCommonAPI=~ m/certificate subject name/m){
		handleErrors(
            "SSL certificate problem:$returnFromCommonAPI"
        );
	}
	else {
		handleErrors(
"Couldn't retrieve data from API. Check the availabilty of Management API and correct credentials. ". $returnFromCommonAPI
		);
	}

}

#Retrieves the links/URLs of XPaths of the same resource type. A type can be a database, appserver,...
#It returns the string after the last slash.
#'/manage/LATEST/servers/Admin?group-id=Default ' --> returns Admin?group-id=Default
#'/manage/LATEST/databases/Documenmts' --> returns Documents
sub getNamesOfResourcesOfASpecificType {
	my ( $resourceType, $returnFromCommonAPI, $link, $verbose ) = @_;
	my $resparser = XML::LibXML->new;
	my $doc       = $resparser->parse_string($returnFromCommonAPI);
	my $fullpath  = "//*[local-name(.)= \"list-items\" ]//*[local-name(.)= \"$link\" ]";
	printVerbose( "-uniqueListOfElements($resourceType): $fullpath",
		VERBOSE3, $verbose );
	my @resources = $doc->findnodes($fullpath);
	my @shortnames;

	foreach my $oneresource (@resources) {
		printVerbose( "-splitNameFromPath:" . $oneresource->textContent,
			VERBOSE3, $verbose );
		push( @shortnames, ( split( '/', $oneresource->textContent ) )[4] );
	}

	printVerbose(
		"-uniqueListOfElements-Result:@shortnames (" . @shortnames . ")",
		VERBOSE3, $verbose );
	return @shortnames;
}

#Verifies that the resource is not excluded.
#It can be excluded by adding an XML element in the service template of the XML input file or by using the global excludes.
#Example for local-exclude in the service-template:
    #<exclude name="Modules"></exclude>
#Returns true if not excluded.
sub isNotExcluded {
	my ( $resource, $verbose, @excludelist ) = @_;
	my $included = 1;
	foreach my $mynode (@excludelist) {
		if ( ref($mynode) =~ m/XML::LibXML::Element/ ) {
			if ( $resource eq $mynode->getAttribute("name") ) {
				$included = 0;
				#excluded
			}
			else {
			}
		}
	}
	printVerbose( "-is excluded:$resource-$included",
		VERBOSE3, $verbose );
	return $included;
}

#instead of using the default threshold it is possibe to have a custom threshold for a resource
#Example: <custom-threshold name="Documents" warning="@2:6" critical="@7:100"></custom-threshold>
sub applyCustomThreshold {
	my ( $elename, $warning, $critical, $verbose, @customthresholds ) = @_;
	foreach my $mynode (@customthresholds) {
		if ( ref($mynode) =~ m/XML::LibXML::Element/ ) {
			if ( $mynode->getAttribute("name") eq $elename ) {
				$warning  = $mynode->getAttribute(WARNING);
				$critical = $mynode->getAttribute(CRITICAL);
				printVerbose(
					"-Applied new theshold to: $elename - $warning-$critical",
					VERBOSE2, $verbose );
			}
		}

	}
	return ( $warning, $critical );
}

#Returns the hostname which will be used for the host_name in the host-template
sub getHostname {
	my ( $unique_name, $resname, $type ) = @_;
	$resname =~ s/\?/-/;
	$resname =~ s/\\\=/-/;
	$resname =~ s/\\\&/-/;
	$resname =~ s/=/-/;      # very important to have this one at the end...
	if ( $type eq HOSTS ) {
		return $resname;
	}
	else {
		return "$unique_name-$resname";
	}
}

# Input: list of resources for one service
#Returns a comma seperated list of host_names.
#Will be used for Nagios service definitions
sub getListOfHostNamesForHostName {
	my ( $uniqueshortname, $verbose, $type, @listofRes ) = @_;
	my $listOfResources = "";
	foreach my $res (@listofRes) {
		if ( length($listOfResources) == 0 ) {
			$listOfResources = getHostname( $uniqueshortname, $res, $type );
		}
		else {
			$listOfResources =
			  $listOfResources . ", "
			  . getHostname( $uniqueshortname, $res, $type );
		}

	}
	if ( length($listOfResources) == 0 ) {
		handleErrors(
            "Error: At least one resource has to be associated with services agains \" $type \". Remove the service definitions from the XML file when it is not needed"
		);
	}
	return $listOfResources;

}

#Returns the macro name for a threshold
#in the format: [WARNING|CRITICAL]-SOMENAME
sub getMacroName {
	my ( $thesholdtype, $servname ) = @_;
	return "$thesholdtype-" . uc $servname;
}

#Returns a macro used in a service template
#It links to the macro defined in the corresponding host template
#Format(based on Nagios custom variables -documentation): $_HOSTVARIABLE$
sub addhostmacro {
	my ($customVariable) = @_;
	return "\$" . "_HOST" . $customVariable . "\$";
}

# Creates a nagios servie defintion.
sub createService {
	my (
		$genericservname, $uniqueshortname, $type,
		$servdes,         $checkcmd,        $servicegroup,
		$withThresholds,  $needMinusW,      $needMinusC,
		$check_interval,  $note,            $verbose,
		@listofRes
	  )
	  = @_;

	my $warningfinal;
	my $criticalfinal;
	if ( $withThresholds == 1 ) {
		my $warningMacro  = getMacroName( "WARNING",  $servdes );
		my $criticalMacro = getMacroName( "CRITICAL", $servdes );
		if ( $needMinusW == 1 ) {
			$warningfinal = "-w " . addhostmacro($warningMacro);
		}
		if ( $needMinusC == 1 ) {
			$criticalfinal = "-c " . addhostmacro($criticalMacro);
		}
	}
	if ($check_interval) {
		$check_interval = "check_interval $check_interval";
	}
	my $listofresources =
	  getListOfHostNamesForHostName( $uniqueshortname, $verbose, $type,
		@listofRes );
	my $servgroupname = getServiceGroupName( $type, $uniqueshortname );

	my $serviceDefString = "\n    define service{
        use            $genericservname
        host_name      $listofresources
        service_description $servdes
        notes           $note
        servicegroups   $servgroupname
        $checkcmd $warningfinal $criticalfinal \$_HOSTMLSSL\$ \$_HOSTMLTIMEOUT\$
        $check_interval
    }
";
	return $serviceDefString;
}

# creates a XML element containing the information to generate a host laster on.
# it generates first all services, later it creates the hosts and includes the thresholding
sub constructXMLElementsNoThreshold {
	my ( $resname, $servdes, $elename, $res ) = @_;

	return constructXMLElements( $resname, $servdes, $elename, 0, undef, undef,
		$res );

}

# executes check_marklogic.pl to validate the check_command.
sub validateCheckCommand{
 my ($checkcmd,$servdesc,$host,$backuphost,$port,$authentication,$verbose,$withThreshold,$timeout,$ssl,$sslcertificate,@listofRes) = @_;

 if(! $checkcmd  =~ m/check_command/){
    handleErrors("check_command is missing at the beginning of this command: $checkcmd");
 }
 if(!defined($timeout))
 {
    $timeout = "--timeout  10";
 }
 else
 {
    $timeout = "";
 }

 $checkcmd =~ s/check_command//;
 # now replace all macros. Do not do any validation here. The check_marklogic.pl plugin will validate the input paramters
 $checkcmd =~ s/\$_HOSTMLUSERPW\$/$authentication/;
 $checkcmd =~ s/\$_HOSTMLPORT\$/$port/;
 $checkcmd =~ s/\$_HOSTMLTIMEOUT\$/$timeout/;
 $checkcmd =~ s/\$_HOSTMLSSL\$/--ssl $ssl/;
 $checkcmd =~ s/\$_HOSTMLSSLCERTIFICATE\$/--sslcertifcate $sslcertificate/;

 $checkcmd =~ s/\$SERVICEDESC\$/$servdesc/;
 $checkcmd =~ s/check_marklogic.pl!/check_marklogic.pl/;
 $checkcmd =~ s/\\\&/&/;
 $checkcmd =~ s/\\\&/&/;
 $checkcmd =~ s/\&/\?/g;

 for my $host ($host,$backuphost)
 {
    if(length($host)>0){
        my $checkcmdtotest = $checkcmd;
        $checkcmdtotest =~ s/\$_HOSTMLIP\$/$host/;
        for my $res (@listofRes){
            my $resreplaced = $checkcmdtotest;
               $resreplaced =~ s/\$_HOSTMLALIAS\$/$res/;

               printVerbose( "-validateCheckCommand: $resreplaced", VERBOSE2, $verbose);
               # in case user specified a threshold, use a dummy threshold to check the exit code of the plugin
               my $dummythreshold = "";
               if($withThreshold eq 1)
               {
                $dummythreshold = "-w 1:1";
               }

               system("perl $resreplaced $dummythreshold >/dev/null");
               my $pluginResponse  = $? >> 8;
               # Nagios plugin code: unkown indicates a problem.
               if ($pluginResponse eq "3"){
                    printVerbose( ";Warning: The service $servdesc against host $host returned Unknown: $resreplaced", VERBOSE1, $verbose);
               }

        }
    }
 }

};

# similar to the one above. Just that this service has a threshold
sub constructXMLElementsWithThreshold {
	my ( $resname, $servdes, $elename, $warning, $critical, $res ) = @_;

	return constructXMLElements( $resname, $servdes, $elename, 1, $warning,
		$critical, $res );
}

# Packages all the elements needed to generate our host later. E.g. it adds the thresholds to the XML data structures.
# <wrapper>
#   <res-name></res-name>
#   <resource-type></resource-type>
#   <alias></alias>
#   <CRITICAL></CRITICAL>
#   <WARNING></WARNING>
# </wrapper>


sub constructXMLElements {
	my (
		$resname, $servdes,  $elename, $withThresholds,
		$warning, $critical, $restype
	  )
	  = @_;

	my $twrapper = XML::LibXML::Element->new("wrapper");
	my $tnode    = XML::LibXML::Element->new("res-name");
	$tnode->appendText($resname);
	my $servd = XML::LibXML::Element->new(SERVICEDESCRIPTION);
	$servd->appendText($servdes);
	my $aliasd = XML::LibXML::Element->new(ALIAS);
	$aliasd->appendText($elename);
	my $resourcetype = XML::LibXML::Element->new(RESOURCETYPE);
	$resourcetype->appendText($restype);

	$twrapper->appendChild($tnode);
	$twrapper->appendChild($servd);
	$twrapper->appendChild($aliasd);
	$twrapper->appendChild($resourcetype);

	if ( $withThresholds == 1 ) {

		my $tc = XML::LibXML::Element->new(CRITICAL);
		$tc->appendText($critical);
		my $tw = XML::LibXML::Element->new(WARNING);
		$tw->appendText($warning);
		$twrapper->appendChild($tc);
		$twrapper->appendChild($tw);
	}
	return $twrapper;

}

#Input Threshold node and either default-critical/ default-warning
#Returns the value of it.
sub getDefaultThreshold {
	my ( $name, $requiresMinusX, $thresholdnode ) = @_;
	my $defaultthresholdnode =
	  ( $thresholdnode->getChildrenByLocalName($name) )[0];
	my $value = undef;
	if ( ref($defaultthresholdnode) =~ m/XML::LibXML::Element/ ) {
		$value          = $defaultthresholdnode->textContent;
		$requiresMinusX = 1;
	}
	return ( $value, $requiresMinusX );
}

sub iterateAndGenerateServices {
	my ( $doc, $resourcetype, $uniqueshortname, $dbcount, $appservercount,$hostcount,$clustercount, $version,
		$verbose,$host,$backuphost,$port,$authentication,$timeout,$ssl,$sslcertificate,
		@uniqResElements )
	  = @_;
	printVerbose( ";-generate Nagios services for group: $resourcetype - See below for a list of services that failed in testing.", VERBOSE1, $verbose );
	printVerbose( ";------------------------------------------", VERBOSE1, $verbose );

	printVerbose( "-iterate over these resources:  @uniqResElements ",
		VERBOSE3, $verbose );
	printVerbose(
"Attention: It might replace parameter markers with those values for thresholds:  $dbcount - $appservercount - $hostcount - $clustercount",
		VERBOSE3, $verbose
	);

	my $serviceDefStrings = "";
	my @wrapperelements;

	my $genericServname =
	  ( $doc->findnodes('//generic-service') )[0]->textContent;
	my $genericHostname = ( $doc->findnodes('//generic-host') )[0]->textContent;
	my $path = '//service-template[@type = \'' . $resourcetype . '\']';
	printVerbose( "-useGenericHost: $genericHostname", VERBOSE3, $verbose );
	printVerbose( "-XPath to retrieve all service templates of the same type: $path ",
		VERBOSE3, $verbose );

	#Generate service templates
	my @resourcetypes = $doc->findnodes($path);
	my $rtype;
	my @managedres;

    #global exludes for resroucce type
    my $path = '//excludes/exclude[@resourcetype = \'' . $resourcetype . '\']';
    my @globalexcludelistforResourceType = $doc->findnodes($path);

	foreach $rtype (@resourcetypes) {

		my $servdes = ( $rtype->getChildrenByLocalName("service_description") );
		my $checkcmd    = ( $rtype->getChildrenByLocalName("check-command") );
		my $threshold   = ( $rtype->getChildrenByLocalName("threshold") )[0];
		my @localexcludelist = ( $rtype->getChildrenByLocalName("exclude") );
		my $note        =
		  ( $rtype->getChildrenByLocalName("service_note") )[0]->textContent;
		my $check_interval = $rtype->getAttribute("refresh");

		if ( $check_interval eq "default" ) {
			$check_interval = undef;
			printVerbose( "-check-interval:default. Service will inherit the refresh interval from the abstract templates.",
        VERBOSE3, $verbose );
		}else{
			printVerbose( "-check-interval:$check_interval",
        VERBOSE3, $verbose );
		}

		my $withThreshold = 0;
		if ( ref($threshold) =~ m/XML::LibXML::Element/ ) {
			$withThreshold = 1;
		}

		printVerbose( "-final-resource type: $resourcetype",      VERBOSE2, $verbose );
		printVerbose( "-final-service-description: $servdes",      VERBOSE2, $verbose );
		printVerbose( "-final-command: $checkcmd",       VERBOSE2, $verbose );
		printVerbose( "-------------------------------", VERBOSE2, $verbose );


	    my @listofRes;
		my $needMinusC = 0;
		my $needMinusW = 0;

		foreach my $elename (@uniqResElements) {
			if (    ( $resourcetype eq $rtype->getAttribute("type") )
				and ( isNotExcluded( $elename, $verbose, @localexcludelist,@globalexcludelistforResourceType ) == 1 ) )
			{
				printVerbose( "-Resource-name:$elename", VERBOSE2, $verbose );
				if ( $withThreshold >= 1 ) {
					my $critical = undef;
					my $warning  = undef;
					printVerbose(
						"-Resource $elename used for service $servdes",
						VERBOSE3, $verbose
					);
					( $warning, $needMinusW ) = getDefaultThreshold(
						"default-warning", $needMinusW,
						$threshold,        $withThreshold
					);
					( $critical, $needMinusC ) = getDefaultThreshold(
						"default-critical", $needMinusC,
						$threshold,         $withThreshold
					);

					my @customthresholds =
					  $threshold->getChildrenByLocalName("custom-threshold");
					( $warning, $critical ) =
					  applyCustomThreshold( $elename, $warning, $critical,
						$verbose, @customthresholds );

					printVerbose(
						"-Final thresholds: $elename ($warning)($critical) ("
						  . $rtype->getAttribute("type") . ")",
						VERBOSE2, $verbose
					);

					push( @listofRes, $elename );

					if ( $warning eq DBCOUNT ) {
						$warning = "$dbcount:$dbcount";
						printVerbose( "-replaceThresholdWithDatabaseCount:$warning",
							VERBOSE3, $verbose );
					}
					if ( $warning eq SERVERCOUNT ) {
						$warning = "$appservercount:$appservercount";
						printVerbose( "-replaceThresholdWithServerCount:$warning",
							VERBOSE3, $verbose );
					}
					if ( $warning eq HOSTCOUNT ) {
                        $warning = "$hostcount:$hostcount";
                        printVerbose( "-replaceThresholdWithHostCount:$warning",
                            VERBOSE3, $verbose );
                    }
                    if ( $warning eq CLUSTERCOUNT ) {
                        $warning = "$clustercount:$clustercount";
                        printVerbose( "-replaceThresholdWitClusterCount:$warning",
                            VERBOSE3, $verbose );
                    }
					my $wrapperelement =
					  constructXMLElementsWithThreshold(
						getHostname( $uniqueshortname, $elename, $resourcetype ),
						$servdes, $elename, $warning, $critical, $resourcetype );
					push( @wrapperelements, $wrapperelement );
				}
				else {
					push( @listofRes, $elename );
					my $wrapperelement =
					  constructXMLElementsNoThreshold(
						getHostname( $uniqueshortname, $elename, $resourcetype ),
						$servdes, $elename, $resourcetype );
					push( @wrapperelements, $wrapperelement );

					#no threshold for service.

				}
				printVerbose( "-ServDesc: $servdes", VERBOSE2, $verbose );

			}

		}

		#here we can generate one service.
		if ( @uniqResElements >= 1 ) {
			$serviceDefStrings = $serviceDefStrings
			  . createService(
				$genericServname,             $uniqueshortname,
				$rtype->getAttribute("type"), $servdes,
				$checkcmd,                    $resourcetype,
				$withThreshold,               $needMinusW,
				$needMinusC,                  $check_interval,
				$note,                        $verbose,
				@listofRes
			  );
			  if($verbose >=1){
			     validateCheckCommand("$checkcmd \$_HOSTMLSSL\$ \$_HOSTMLSSLCERTIFICATE\$ \$_HOSTMLTIMEOUT\$",$servdes,$host,$backuphost,$port,$authentication,$verbose,$withThreshold,$timeout,$ssl,$sslcertificate,@listofRes);
		      }
		}
		else {
			$serviceDefStrings = $serviceDefStrings
			  . createService(
				$genericServname,             $uniqueshortname,
				$rtype->getAttribute("type"), $servdes,
				$checkcmd,                    $resourcetype,
				$withThreshold,               $needMinusW,
				$needMinusC,                  $check_interval,
				$note,                        $verbose,
				CLUSTER
			  );
			 if($verbose >=1){
			     validateCheckCommand("$checkcmd \$_HOSTMLSSL\$ \$_HOSTMLSSLCERTIFICATE\$ \$_HOSTMLTIMEOUT\$",$servdes,$host,$backuphost,$port,$authentication,$verbose,$withThreshold,$timeout,$ssl,$sslcertificate,@listofRes);
			 }
		}

	}

	return $genericHostname, $serviceDefStrings, @wrapperelements;
}

sub getAlias {
	my ($alias) = @_;
	return $alias;
}

sub getHostGroupName {
	my ($hostgroup) = @_;
	return $hostgroup;

}

sub createHostGroup {
	my ( $hostgroup, $shortname, $verbose ) = @_;
	my $hostgroupName = getHostGroupName($hostgroup);
	my $alias         = getAlias($shortname);
	printVerbose( "-createHostGroup:$hostgroupName # $alias",
		VERBOSE3, $verbose );
	my $hostgroupDefString = "\n    define hostgroup{
        hostgroup_name " . $hostgroupName . "
        alias   " . $alias . "
    }
    ";
	return $hostgroupDefString;
}

sub getServiceGroupName {
	my ( $servicegroup, $shortname ) = @_;
	return "$shortname-$servicegroup";
}

sub createServiceGroup {
	my ( $servicegroup, $shortname ) = @_;
	return "\n    define servicegroup{
        servicegroup_name "
	  . getServiceGroupName( $servicegroup, $shortname ) . "
        alias   " . ($shortname) . "
    }
";
}

sub getUniqueResourceTypes {
	my ( $verbose, @resourcetypes ) = @_;
	my @managedres;
	foreach my $rtype (@resourcetypes) {
		my $name = $rtype->getAttribute("type");
		push( @managedres, $name );
	}
	@managedres = uniq @managedres;
	printVerbose( "-getUniqueResourceTypes:@managedres", VERBOSE3, $verbose );
	return uniq @managedres;
}

sub formatThreshold {
    my ( $name, $servd, $threshold ) = @_;
    $name = lc $name;
    my $returnString = undef;

           if ( length($threshold) > 0 ) {
                $returnString = "        _"
                  . ( uc $name ) . "-"
                  . ( uc $servd )
                  . " $threshold \n";
            }
    return $returnString;
}

sub getThresholdDefForHostDef {
    my ( $name, $servd, $node ) = @_;
    $name = lc $name;
    my $returnString = undef;

    my $tempnode = ( $node->getChildrenByLocalName($name) )[0];
    if ( ref($tempnode) ) {

        if ( ref($node) ) {
            my $threshold =
              ( $node->getChildrenByLocalName($name) )[0]->textContent;
            if ( length($threshold) > 0 ) {
                $returnString = $threshold;
            }
        }

    }
    return $returnString;

}

sub printHelp {
	my ($verbose) = @_;
	printVerbose( "Starting Help", VERBOSE3, $verbose );
	printVerbose( "Help - generate_marklogic_config.pl",
		OUTPUTTOSHELL, $verbose );
	printVerbose( "-----------------------------", OUTPUTTOSHELL, $verbose );
	printVerbose(
"This script connects to the Management API of MarkLogic server and retrieves the name of all defined resources.",
		OUTPUTTOSHELL, $verbose
	);
	printVerbose(
"The XML input file defines all services. Based on those and the available resources, it generates a Nagios config file.",
		OUTPUTTOSHELL, $verbose
	);
	printVerbose( "-----------------------------", OUTPUTTOSHELL, $verbose );
	printVerbose( "-a (--authentication) <username:password>",
		OUTPUTTOSHELL, $verbose );
	printVerbose( "-H (--host)           <hostname> ", OUTPUTTOSHELL,
		$verbose );
	printVerbose( "-p (--port)           <port>             (e.g. 8002)",
		OUTPUTTOSHELL, $verbose );
	printVerbose(
		"-f (--filename)       <filename.xml>     (e.g. ml_input_template.xml)",
		OUTPUTTOSHELL, $verbose
	);
	printVerbose( "-c (--clustername)    <string>           (e.g. My-Cluster1",
		OUTPUTTOSHELL, $verbose );
	printVerbose( "-u (--uniqueshortcut) <string>           (e.g. Clu-1) ",
		OUTPUTTOSHELL, $verbose );
	printVerbose("---------+ Optional parameters ------------------",
		OUTPUTTOSHELL, $verbose
	);
	printVerbose( "-verbose              0|1|2|3 ",
        OUTPUTTOSHELL, $verbose );
    printVerbose( "-t (--timeout)        <timeout-seconds>",
        OUTPUTTOSHELL, $verbose );
    printVerbose( "-ssl (--ssl)          <0|1> (default 0)",
        OUTPUTTOSHELL, $verbose );
    printVerbose( "-cert (--sslcertificate)          <fullpath> (requires -ssl 1)",
        OUTPUTTOSHELL, $verbose );
 	#printVerbose( "-bh (--backuphost)    <alternative-hostname>",
    #    OUTPUTTOSHELL, $verbose );

    printVerbose( "---------OR------------------", OUTPUTTOSHELL, $verbose );
    printVerbose( "-h (--help)",                   OUTPUTTOSHELL, $verbose );
    printVerbose( "-V (--version) ",               OUTPUTTOSHELL, $verbose );

	printVerbose(
		"---------How to customize the XML input file------------------",
		OUTPUTTOSHELL, $verbose );
	printVerbose(
"1) Modify the refresh interval: Modify the attribute 'refresh'. Specify 'default' to use the setting in the generic template. ",
		OUTPUTTOSHELL, $verbose
	);
	printVerbose(
"2) Want to add a threshold for a service? Add such an node in your XML file:  <threshold> <default-critical>@10:399</default-critical> <default-warning>@400:1000</default-warning></threshold> ",
		OUTPUTTOSHELL, $verbose
	);
	printVerbose(
"3) Want to set a custom threshold: Add such an element in the threshold element. <custom-threshold name=\"MyDatabase\" warning=\"@5000:7000\" critical=\"@7001:10000\"></custom-threshold>",
		OUTPUTTOSHELL, $verbose
	);
	printVerbose(
"4) Want to exclude/ not to monitor a resource for a certain service (local exclude)? Add such an node in the \"service\"-element:<exclude name=\"MyDatabase\"></exclude>",
		OUTPUTTOSHELL, $verbose
	);
    printVerbose(
"5) Want to exclude/ not to monitor a resource totally (global exclude)? Add such an node in the \"excludes\" element:<exclude resourcetype=\"Databases\" name=\"App-Services\" />",
        OUTPUTTOSHELL, $verbose
    );
	printVerbose(
"6) Want to debug a service out of Nagios? add: --verbose 1 in the command description. ",
		OUTPUTTOSHELL, $verbose
	);
	printVerbose( "",         OUTPUTTOSHELL, $verbose );
	printVerbose( "Example:", OUTPUTTOSHELL, $verbose );
	printVerbose(
"perl generate_marklogic_config.pl -a user:pw --host localhost  -port 8002  -verbose 0 -clustername MyCluster -filename ml_input_template.xml -u ML-1",
		OUTPUTTOSHELL, $verbose
	);
}

sub getXMLElementNameForResource {
	my ( $resourceName, $linkSequence, $verbose ) = @_;
	my $parser          = XML::LibXML->new;
	my $doc             = $parser->parse_string($linkSequence);
	my $xpath           = "//link[\@resourcetype = \"$resourceName\"]";
	my $resourceElement = ( $doc->findnodes($xpath) );
	printVerbose(
		"-getXMLElementNameForResource:$resourceName-$resourceElement",
		VERBOSE3, $verbose );
	if ( length($resourceElement) <= 0 ) {
		handleErrors(
			"Could not find XML element for resource type:$resourceName");
	}
	return $resourceElement;
}

sub getAPIName{
    my ( $resourceName, $linkSequence, $verbose ) = @_;
    my $parser          = XML::LibXML->new;
    my $doc             = $parser->parse_string($linkSequence);
    my $xpath           = "//link[\@resourcetype = \"$resourceName\"]";
    my $apiName = ( $doc->findnodes($xpath))[0]->getAttribute("apiName");

    printVerbose(
        "-getAPIName:$resourceName-$apiName",
        VERBOSE3, $verbose );
    return $apiName;
}

sub getResourceFilter {
    my ( $resourceName, $linkSequence, $verbose ) = @_;
    my $parser          = XML::LibXML->new;
    my $doc             = $parser->parse_string($linkSequence);
    my $xpath           = "//link[\@resourcetype = \"$resourceName\"]";
    my $resourceFilter = ( $doc->findnodes($xpath))[0]->getAttribute("resource-filter");

    printVerbose(
        "-getResourceFilter:$resourceName-$resourceFilter",
        VERBOSE3, $verbose );
    return $resourceFilter;
}

sub getConditionWithResroucesToIterate {
    my ( $linkSequence, $toiterate, $verbose ) = @_;
    my $parser          = XML::LibXML->new;
    my $doc             = $parser->parse_string($linkSequence);
    my $xpath           = "//link[ \@iterate = \"" .$toiterate . "\"]/\@resourcetype";
    my @resourceTypes = ( $doc->findnodes($xpath) );
    my $xpathcondition = undef;
    foreach (@resourceTypes)
    {
    	if(not defined $xpathcondition){
    	   $xpathcondition = "\@type=\"".  $_->textContent("resourcetype") . "\"";
    	}else{
            $xpathcondition	= "$xpathcondition" . " or "  . "\@type=\"".  $_->textContent("resourcetype") . "\"";
    	}
    }
    printVerbose("-XPathCondition: $xpathcondition",
        VERBOSE2, $verbose );
    return $xpathcondition;

}

sub getSSLSettingsForAbStractHost{
	my ( $ssl, $sslcertificate ) = @_;
	my $SSLstring = "_MLSSL";

	if(length($ssl) >0){
        $SSLstring= "$SSLstring --ssl $ssl";
        if(length($sslcertificate) >0){
            $SSLstring= "$SSLstring --sslcertificate $sslcertificate";
        }
    }
    return $SSLstring;
}

sub getTimeoutSettingsForAbStractHost{
    my ( $timeout ) = @_;
    my $timeoutstring = "_MLTIMEOUT";

    if(length($timeout) >0){
        $timeoutstring= "$timeoutstring --timeout $timeout";
    }
    return $timeoutstring;
}

sub getHostCheckCommand {
	my ( $restype, $alias ) = @_;
	my $hostcheckcommand = undef;
	my $url              = undef;
	my $key              = undef;
	my $value            = undef;
	my $critical         = undef;
	if ( $restype->textContent eq DBS ) {
		$url              = "databases/$alias?view=status";
		$key              = "state";
		$value            = "available";
		$critical         = "\@0:0";
		$hostcheckcommand =
"check_command check_marklogic.pl!  -a \$_HOSTMLUSERPW\$  -port \$_HOSTMLPORT\$  --host \$_HOSTMLIP\$ \$_HOSTMLSSL\$ \$_HOSTMLTIMEOUT\$ -path /manage/$apiversion/$url -key $key -c $critical -op eq=$value";
	}
	elsif ( $restype->textContent eq HOSTS ) {
		$url              = "hosts/$alias?view=status";
		$key              = "online";
		$value            = "true";
		$critical         = "\@0:0";
		$hostcheckcommand =
"check_command check_marklogic.pl!  -a \$_HOSTMLUSERPW\$  -port \$_HOSTMLPORT\$  --host \$_HOSTMLIP\$ \$_HOSTMLSSL\$ \$_HOSTMLTIMEOUT\$ -path /manage/$apiversion/$url -key $key -c $critical -op eq=$value";
	}
	elsif ( $restype->textContent eq APPSERS ) {
		$url = "servers/$alias?view=status";
		$key              = "enabled";
		$value            = "true";
		$critical         = "\@0:0";
		$hostcheckcommand =
"check_command check_marklogic.pl!  -a \$_HOSTMLUSERPW\$  -port \$_HOSTMLPORT\$  --host \$_HOSTMLIP\$ \$_HOSTMLSSL\$ \$_HOSTMLTIMEOUT\$ -path /manage/$apiversion/$url -key $key -c $critical -op eq=$value";
	}
	elsif ( $restype->textContent eq CLUSTER ) {
		$hostcheckcommand = "check_command check_ML_ping!  \$_HOSTMLIP\$ ";
	}
	else {

		#defined your own resource? add it here.
		$hostcheckcommand =
		  "; no check-command available for " . $restype->textContent;
	}
	return $hostcheckcommand;
}

# Step 1a
# Parse Input args
my %inputArgs = ();
local $@;
eval {
	local $SIG{__WARN__};

	#uses Getop:Long
	%inputArgs = parseInput($ARGV);
};

if ( exists $inputArgs{version} ) {
	printVerbose( "generate_marklogic_config.pl: version 1.0.1",
		OUTPUTTOSHELL, $inputArgs{verbose} );
	exit 0;
}
elsif ( exists $inputArgs{help} ) {
	printHelp();
	exit 0;
}

printInput( $inputArgs{verbose}, ";", %inputArgs );

checkDefaultParameters(%inputArgs);
checkOptionalParameters(%inputArgs);

# Step 1b
# determine version

my $adminuserpw = $inputArgs{adminuserpw} ? $inputArgs{adminuserpw} : $inputArgs{authentication};

my $check_version = "curl \'http://$inputArgs{host}:8002/manage/v1/\' --user $adminuserpw -i --silent --anyauth -H \'Accept: application/xml\'";

my $returnFromCommonAPI = `$check_version`;

if ( $returnFromCommonAPI =~  m/HTTP\/1.1 200/ ) {
    $apiversion = "v1";
}else{
    $apiversion = "v2";
}

# Step 2
# Open config file
my $finalConfigString = "";
my $linkSequence = getLinkSequence();

# create group for this cluster
$finalConfigString =
  createHostGroup( $inputArgs{clustername}, $inputArgs{uniqueshortcut},
	$inputArgs{verbose} );

my @uniqueListOfResoureceTypes;
my @wrapperelements;
my $genericHostname;

my $filename;

if ($inputArgs{filename}){
    #ok
    $filename = $inputArgs{filename}
}else{
    # no -f supplied, determine from apiversion

    if($apiversion =~ "v2"){
        $filename = ML7_CONFIG_TEMPLATE;
    }
    else{
        $filename = ML56_CONFIG_TEMPLATE;
    }

    # used in development
    $filename = -e $filename ? $filename : "../XMLConfigFileForGeneratingConfigs/".$filename;
}

my $parser           = XML::LibXML->new;
my $doc              = $parser->parse_file( $filename );
my $generictemplates = ( $doc->findnodes('//generic-templates') )[0];

#there is just one. containing the names of the abstract templates which a user wants to use for all of their services and hosts
#makes it possible to have different polling intervals,...

@uniqueListOfResoureceTypes = getUniqueResourceTypes(
    $inputArgs{verbose},
    $doc->findnodes(
        '//service-template[' . getConditionWithResroucesToIterate($linkSequence,"true", $inputArgs{verbose}) . ']')
);


my $tempServiceDefString = undef;
my $dbcount              = undef;
my $appservercount       = undef;
my $hostcount            = undef;
my $clustercount         = undef;

# databases, hosts,... (everything what has subresources)
foreach my $resourceType (@uniqueListOfResoureceTypes) {
	$finalConfigString =
	  $finalConfigString
	  . createServiceGroup( $resourceType, $inputArgs{uniqueshortcut} );

	my $link =
	  getXMLElementNameForResource( $resourceType, $linkSequence,
		$inputArgs{verbose} );
    my $resourcefilter = getResourceFilter($resourceType,$linkSequence,$inputArgs{verbose} );
    my $apiName = getAPIName($resourceType,$linkSequence,$inputArgs{verbose} );

    #get all resources from the API
	my $url = constructCurlCommand(
		$inputArgs{host},                          $inputArgs{port},
		"/manage/$apiversion/" . lc $apiName. $resourcefilter, $inputArgs{authentication},
		$inputArgs{timeout},                       $inputArgs{ssl},
		$inputArgs{sslcertificate},
		$inputArgs{verbose}
	);
	my $returnFromCommonAPI = `$url`;
	$returnFromCommonAPI =
	  checkheadersandreturnXML( $returnFromCommonAPI, $inputArgs{verbose} );
	my @uniqlistofElements =
	  getNamesOfResourcesOfASpecificType( $resourceType, $returnFromCommonAPI,
		$link, $inputArgs{verbose} );

#now go back to the config file and look for all the service templates for the current resource type
	my @temp;

    # this script uses the counts of resources for thresholds. This is very limited. See below.
	if ( $resourceType eq DBS ) {
		$dbcount = @uniqlistofElements;
		printVerbose( "-setDBCount:$dbcount", VERBOSE3, $inputArgs{verbose} );
	}
	if ( $resourceType eq APPSERS ) {
		$appservercount = @uniqlistofElements;
		printVerbose( "-setAppServerCount:$appservercount",
			VERBOSE3, $inputArgs{verbose} );
	}
	if ( $resourceType eq HOSTS ) {
        $hostcount = @uniqlistofElements;
        printVerbose( "-setHostCount:$hostcount",
            VERBOSE3, $inputArgs{verbose} );
    }
    if ( $resourceType eq CLUSTERS ) {
        $clustercount = @uniqlistofElements;
        printVerbose( "-setClusterCount:$clustercount",
            VERBOSE3, $inputArgs{verbose} );
    }
    if(@uniqlistofElements >0 ){
        ( $genericHostname, $tempServiceDefString, @temp ) =
    	  iterateAndGenerateServices( $doc, $resourceType,
    		$inputArgs{uniqueshortcut},
    		$dbcount, $appservercount, $hostcount,$clustercount, $apiversion, $inputArgs{verbose},$inputArgs{host},$inputArgs{backuphost},$inputArgs{port},$inputArgs{authentication},
    		$inputArgs{timeout},$inputArgs{ssl},$inputArgs{sslcertificate},@uniqlistofElements);

    	$finalConfigString = $finalConfigString . $tempServiceDefString;
    	push( @wrapperelements, @temp );
    }
}

my @uniqueListOfResoureceTypes;

# for cluster . no sub resources
my @uniqueListOfResoureceTypes  = getUniqueResourceTypes(
    $inputArgs{verbose},
    $doc->findnodes(
                '//service-template[' . getConditionWithResroucesToIterate($linkSequence,"false", $inputArgs{verbose}) . ']'));

#@uniqueListOfResoureceTypes = uniq @uniqueListOfResoureceTypes;
foreach my $res (@uniqueListOfResoureceTypes) {
	$finalConfigString =
	  $finalConfigString
	  . createServiceGroup( $res, $inputArgs{uniqueshortcut} );

	my @temp;
    if(@uniqueListOfResoureceTypes >0 ){
    	( $genericHostname, $tempServiceDefString, @temp ) =
    	  iterateAndGenerateServices( $doc, $res, $inputArgs{uniqueshortcut},
    		$dbcount, $appservercount, $hostcount,$clustercount, $apiversion, $inputArgs{verbose},$inputArgs{host},$inputArgs{backuphost},$inputArgs{port},$inputArgs{authentication},
    		$inputArgs{timeout},$inputArgs{ssl},$inputArgs{sslcertificate},
    		@uniqueListOfResoureceTypes );
    	$finalConfigString = $finalConfigString . $tempServiceDefString;
    	push( @wrapperelements, @temp );
    }
}

my $root = XML::LibXML::Document->new('document');
my $node = XML::LibXML::Element->new("docroot");

foreach my $n (@wrapperelements) {
	if ( ref($n) =~ m/XML::LibXML::Element/ ) {
		$node->appendChild($n);
	}
}

$root->setDocumentElement($node);
my @nodes =
  $root->findnodes(
'/docroot[not( wrapper/res-name/text() = preceding-sibling::wrapper/res-name/text())]/wrapper/res-name/text()'
  );
if ( @nodes == 0 ) {
	handleErrors("Couldn't retrieve XML nodes from XML file");
}
my @distinctres;
foreach my $resource (@nodes) {
	push( @distinctres, $resource->textContent );
}
@distinctres = uniq @distinctres;

my $hostforcredentials = "$inputArgs{uniqueshortcut}_abstract";

# Add abstract e-node.

my $backuphosts = $inputArgs{backuphost};
    if (length($backuphosts)>1){
        $backuphosts = ",$backuphosts";
    }
$finalConfigString = $finalConfigString . "\n     define host{
        use     $genericHostname
        name   $hostforcredentials
        _MLPORT $inputArgs{port} ;
        _MLUSERPW $inputArgs{authentication};
        _MLIP $inputArgs{host}$backuphosts;
        "
        .
         getSSLSettingsForAbStractHost($inputArgs{ssl},$inputArgs{sslcertificate})
        .
        "
        "
        .
         getTimeoutSettingsForAbStractHost($inputArgs{timeout})
        .
        "
        register 0;
    }
";

printVerbose( "-------------------------------",
    VERBOSE3, $inputArgs{verbose} );
# Generate all host definitions, each host will inherit the credentials from the e-node (abstract host) above.
printVerbose( "-generateHostResources:" . @distinctres,
	VERBOSE2, $inputArgs{verbose} );
foreach my $longhostname (@distinctres) {
	printVerbose( "-generateHost: $longhostname", VERBOSE2, $inputArgs{verbose} );
	my $path  = "/docroot/wrapper[res-name/text() = '" . $longhostname . "']";
	my @nodes = $root->findnodes($path);
	my $alias        = undef;
	my $thresholds   = "";

	my $hostcheckcommand = "";
	foreach $node (@nodes) {
		$alias = ( $node->getChildrenByLocalName(ALIAS) )[0]->textContent;
		my $servd =
		  ( $node->getChildrenByLocalName(SERVICEDESCRIPTION) )[0]
		  ->textContent;
		$hostcheckcommand =
		  getHostCheckCommand( ( $node->getChildrenByLocalName(RESOURCETYPE) )[0],
			$alias );

     # there can be warning or critical elements, but they don't have to be there. based on the service description

        my $warning =  getThresholdDefForHostDef( WARNING, $servd, $node );
        my $critical = getThresholdDefForHostDef( CRITICAL, $servd, $node );

        #very late, but this is the only place where all thresholds are available for validation.
        my $threshold = Nagios::Monitoring::Plugin::Threshold->set_thresholds(
            warning  => "$warning",
            critical => "$critical"
        );


		$thresholds =
		  $thresholds . formatThreshold( WARNING, $servd, $warning );
		$thresholds =
		  $thresholds . formatThreshold( CRITICAL, $servd, $critical );
	}

	my  $alias_replaced = undef;
	$alias = "_MLALIAS              $alias";
	if($alias =~ m/\&/){
	   # for all resources which have a & in their name, add a second Alias which contains a ? instead.
	   $alias =~ s/\&/\?/;
    }

	$finalConfigString = $finalConfigString . "
    define host{
        use                     $hostforcredentials
        host_name               $longhostname
        address                 $longhostname
        hostgroups              $inputArgs{clustername}
        $hostcheckcommand
        $alias

$thresholds
    }
    ";

}

#print the generated configuration.
printVerbose( "\n ###################\n ", VERBOSE2, $inputArgs{verbose} );
print "\n $finalConfigString";
